#include <fstream>
#include <string>
#include <algorithm>
#include <cmath>
#include <iostream>
#include <memory.h>
#include "cone.h"
#include "basics.h"
#include "event.h"
#include <sstream> 
#include <utility>

/// Boost a particle by a 4-vector. 
void SingleParticle::bst(const Vec4 & pBoost){
	  
  p.bst(pBoost);
  pT  = p.pT();
  y   = p.eta();
  phi = p.phi();	  
}

//--------------------------------------------------------------------------
/// Add energy deposition to cells hit by particles in Event.
vector<SingleParticle> const Cone::addEnergyDeposition(const Event& event) const {

  vector<SingleParticle> my_cells;	
  for (int i = 0; i < event.size(); ++i) {
   // Find particle position in (eta, phi, pT) space.
    double etaNow = event[i].eta();
//    if (abs(etaNow) > etaMax) continue;
  //  double phiNow = event[i].phi();
    double pTnow  = event[i].pT();
  

    if (abs(etaNow) < etaMax && pTnow > pTjetMin) {
      
      my_cells.push_back( SingleParticle(event[i]) );
    }
  
}

	
return my_cells;	
	

}
 
 
/// Analyze event and return a pair <peak template, max overlap>.

pair<vector<SingleParticle>, double> Cone::analyze(const Event& event, const vector<SingleParticle> & comparison_template, double R0) 
{
 jets.clear();
 double jetRapidity = event.axis().eta() ;	  
 double jetPhi = event.axis().phi()  ;
 
 //CHECK delta Phi below!!!
 for (int i = 0; i < event.size(); ++i) {	
  double etaNow = event[i].eta();
  double phiNow = event[i].phi();
  double pTnow = event[i].pT();
  double mNow = event[i].mCalc(); 
  Vec4 pNow(pTnow * cos(phiNow), pTnow * sin(phiNow), pTnow * sinh(etaNow), 
  sqrt( mNow * mNow + pTnow * pTnow * cosh(etaNow) * cosh(etaNow)) );
  jets.push_back( SingleParticle(pNow, pTnow, event[i].eta() - jetRapidity, phiNow - jetPhi) );
}
    
 double maxOverlap = 0.;
 int jmax =0;
 all_overlaps.clear();

 for (int i = 0; i < int (templates.size()); ++i){

  vector<SingleParticle> my_template = templates[i];

if (R0 >0.001) {if (overlapDistance(comparison_template, my_template, R0)< 0.8) continue;}
		double overlapNow = Mode ? GetOverlapGaussian(my_template) : GetOverlap(my_template);
		all_overlaps.push_back(overlapNow);
		if (overlapNow > maxOverlap){ jmax = i; maxOverlap = overlapNow;}
 }
 return make_pair(templates[jmax], maxOverlap);
}


/// Load templates from file.
int Cone::loadTemplates(ifstream& is) {

  string line, tag;
 
 //Event particlesSave;
 
vector<SingleParticle> particlesSave; 
templates.clear();


//Read templates one by one, 
//Format: 
//  nup         # number of particles in template
//  pz py pz E  # 4-vector of particle 1
//  pz py pz E  # 4-vector of particle 2
//    ............................
//  pz py pz E  # 4-vector of particle N


  while (getline(is,line)){
  istringstream getpro(line);
  int nupSave;
  getpro >> nupSave;
  if (!getpro) return false;

  // Reset particlesSave vector.
  particlesSave.clear(); 

  // Read in particle info one by one, and store it.
  double pup1, pup2, pup3, pup4;
  
   if (!getline(is, line)) return false;
    istringstream getall(line);
    getall  >> pup1 >> pup2 >> pup3 >> pup4;
  
  for (int ip = 0; ip < nupSave; ++ip) { 
    if (!getline(is, line)) return false;
    istringstream getall(line);
    getall  >> pup1 >> pup2 >> pup3 >> pup4;
    if (!getall) return false;   
    
   Vec4 pTemp(pup1, pup2, pup3, pup4);


   particlesSave.push_back(SingleParticle(pTemp, pTemp.pT(), pTemp.y(), pTemp.phi()));
  }
 
 if (CheckTemplate (particlesSave) || Mode) templates.push_back(particlesSave);

		
 
  // Reading worked.
  
}
  return (templates.size());

}


//Listing templates
void Cone::list_templates(ostream& os, int Nmax) {
  for (int i=0; i< Nmax;++i )
    { cout << templates[i].size()<< endl;
	Vec4 pSum(0.);
	for (int j=0; j< int(templates[i].size());++j) pSum +=templates[i][j].p;
	cout << pSum;
	for (int j=0; j< int(templates[i].size());++j)
	cout << templates[i][j].p;
}
}

bool Cone::CheckTemplate (const vector<SingleParticle> & probe_template)
	  
	  {  

	bool isOK = true; 
	  for (int i = 0 ; ((i < int (probe_template.size())) && isOK) ; ++i)  
			{  for (int j = i + 1 ; j < int (probe_template.size()); ++j)
				{	
					if( probe_template[i].deltaR2(probe_template[j]) < 4 * subConeRadius * subConeRadius)
				{ isOK = false; break ;}
}
  } 
	  return (isOK);
	  }


//----------------------------------------------------------------------
/// return squared cylinder (eta-phi) distance between this jet and another one.
double SingleParticle::deltaR2(const SingleParticle & other) const {
	
	double  dPhi = abs(phi - other.phi );
      if (dPhi > M_PI) dPhi = 2. * M_PI - dPhi;
    double  dEta =  y - other.y;
      return (dEta * dEta + dPhi * dPhi );
	
}

///Reset jet flags.
void Cone::reset(){
for (int j = 0; j < int(jets.size()); ++j) 
      if (jets[j].isUsed) jets[j].isUsed = false;	
        
	}

/// Sum pT around template momenta and return overlap.

double Cone::GetMassInsideCone (const vector<SingleParticle> & probe_template) {
Vec4 pNow;
    for(int k = 0; k < int(probe_template.size()); ++k) {
       //  Sum up unused particles within required distance of parton.
    for (int j = 0; j < int(jets.size()); ++j) {
        
    if (jets[j].deltaR2(probe_template[k]) > subConeRadius * subConeRadius ) continue;
   	 pNow += jets[j].p;
      
    }


}    
// Return mass.
return (pNow.mCalc());
    }


/// Sum pT around template momenta and return overlap.

double Cone::GetOverlap (const vector<SingleParticle> & probe_template) {	
double sum = 0.;
    for(int k = 0; k < int(probe_template.size()); ++k) {
		    double eTjetNow = 0.;
	   //  Sum up unused particles within required distance of parton.
    for (int j = 0; j < int(jets.size()); ++j) {
      if (jets[j].isUsed) continue;
	if (jets[j].deltaR2(probe_template[k]) > subConeRadius * subConeRadius ) continue;
         eTjetNow += jets[j].pT;
	    jets[j].isUsed = true;
    }
    double eTcenterNow = probe_template[k].pT;
    //change sigma to E/3
    sum += 4.5 * (eTjetNow - eTcenterNow) * (eTjetNow - eTcenterNow)/(eTcenterNow * eTcenterNow); 
    
            Vec4 pNow(0.) ;
	    
	//if one overlap is too low, we skip the rest. 
    if (sum > 4.6 ) break;
} 
// Clean flags. 
reset();
// Return overlap. 
return (exp(-sum));
	}



/// Sum pT around template momenta and return overlap.                  

double Cone::overlapDistance (const vector<SingleParticle> & jet01, const vector<SingleParticle> & jet02, double R0) {
double sum = 0.;
vector<SingleParticle> jet1= jet01,jet2=jet02;
    for(int k = 0; k < int(jet1.size()); ++k) {
            double eTjetNow = 0.;              
       //  Sum up unused particles within required distance of parton.
    for (int j = 0; j < int(jet2.size()); ++j) {
      if (jet2[j].isUsed) continue;
    if (jet2[j].deltaR2(jet1[k]) > R0 * R0 ) continue;
         eTjetNow += jet2[j].pT;
        jet2[j].isUsed = true;
    }              
    double eTcenterNow = jet1[k].pT;
    //change sigma to E/3
    sum += 4.5 * (eTjetNow - eTcenterNow) * (eTjetNow - eTcenterNow)/(eTcenterNow * eTcenterNow);

            Vec4 pNow(0.) ;
    if (sum > 4.6 ) break;
}
//Reset flags

for (int j = 0; j < int(jet2.size()); ++j)    
      if (jet2[j].isUsed) jet2[j].isUsed = false;


// Return overlap. 
return (exp(-sum));
    }



/// Sum pT around template momenta and return overlap.

double Cone::GetOverlapGaussian (const vector<SingleParticle> & probe_template) {   

vector<double> eTjetNow(probe_template.size(),0.);
for (int j = 0; j < int(jets.size()); ++j) {
  // find closest parton
  int jmin = 0;
  double dmin = jets[j].deltaR2(probe_template[0]);
  for(int k = 1; k < int(probe_template.size()); ++k) {
  double dnow = jets[j].deltaR2(probe_template[k]);
  if (dnow < dmin ){dmin = dnow; jmin = k;}
}
double etaJet = jets[j].y - probe_template[jmin].y;
double phiJet = abs(jets[j].phi - probe_template[jmin].phi); 
if (phiJet > M_PI) phiJet = 2. * M_PI - phiJet;
eTjetNow[jmin] += jets[j].pT * exp (-( etaJet * etaJet + phiJet * phiJet)/(2 * sigma * sigma)   ) ;
		 
		 }
		 
double sum = 0.;
for(int k = 0; k < int(probe_template.size()); ++k) {
  double eTcenterNow = probe_template[k].pT;
  sum += 4.5 * (eTjetNow[k] - eTcenterNow) * (eTjetNow[k] - eTcenterNow)/(eTcenterNow * eTcenterNow); }
return (exp(-sum));
}


double Cone::GetOverlapGaussian2 (const vector<SingleParticle> & probe_template) {   

vector<double> eTjetNow(probe_template.size(),0.);
for (int j = 0; j < int(jets.size()); ++j) {
  // find closest parton
  int jmin = 0;
  double dmin = jets[j].deltaR2(probe_template[0]);
  for(int k = 1; k < int(probe_template.size()); ++k) {
  double dnow = jets[j].deltaR2(probe_template[k]);
  if (dnow < dmin ){dmin = dnow; jmin = k;}
}
double etaJet = jets[j].y - probe_template[jmin].y;
double phiJet = abs(jets[j].phi - probe_template[jmin].phi); 
if (phiJet > M_PI) phiJet = 2. * M_PI - phiJet;
eTjetNow[jmin] += jets[j].pT * exp (-( etaJet * etaJet + phiJet * phiJet)/(2 * sigma * sigma)   ) ;
		 
		 }
		 
double sum = 0.;
for(int k = 0; k < int(probe_template.size()); ++k) {
  double eTcenterNow = probe_template[k].pT;
  sum += 4.5 * (eTjetNow[k] - eTcenterNow) * (eTjetNow[k] - eTcenterNow)/(eTcenterNow * eTcenterNow); }
return (exp(-sum));
}

