// event.h is a part of the TEMPLATE Higgs tagger.

// Header file for the Event class.
// Event: simple container for the particles in a jet or a template.

#ifndef Event_H
#define Event_H

#include <cmath>
#include <string>
#include <vector>
#include "basics.h"
#include <algorithm>

#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"



using namespace std;


bool sortVecByPt(const Vec4 &d1, const Vec4 &d2);




/// The basic class for holding the jet constituents and the particles in a template. It also computes basic jet shapes, such as planar flow.

class Event {
 public:
 Event() {};
 bool readEvent(std::ifstream& in );
bool readEvent(fastjet::PseudoJet& myJets);
//bool readEvent(vector<SingleParticle> & my_particles) ;


 // Overload index operator to access element of event record.
 Vec4& operator[](int i) {return particles[i];}
 const Vec4& operator[](int i) const {return particles[i];}
 const Vec4& axis() const {return jet_axis;}
 
 // sort by age ascending (low to high)
 void sortByPt() {sort(particles.begin(), particles.end(), sortVecByPt);}
   
 // Event record size.
  int size() const {return particles.size();}
  
 double planarFlow() const;
 double planarFlow2(double r = 0.) const;
 double angularity(int a, double R) const;
 double thetaS() const;
 double thetaBar() const;
double dipolarity(Vec4 j1, Vec4 j2) const;
  double dipolarity2(Vec4 j1, Vec4 j2) const;


  
  // Clear event record.
  void clear() {particles.clear();jet_axis.reset();}

  // Put a new particle at the end of the event record; return index.
  
  int append(Vec4 entryIn) {    
    particles.push_back(entryIn);
	jet_axis += entryIn;  
    return particles.size() - 1;
  }
  
  void bst(const Vec4& vec) 
    {for (int i = 0; i < size(); ++i) particles[i].bst(vec);}
  
 private:
 
   vector<Vec4> particles;
   Vec4 jet_axis;
   
   
   // Constants: could only be changed in the code itself.
  static const double TINY;

};


#endif // end Event_H
